from __future__ import annotations

from typing import Any, Collection, Mapping

from typing_extensions import Self

from ._patches import Omit, Only, Patched

__all__ = ["Omit", "Only", "Patched"]


class _Predicate:
    def __call__(self) -> Self:
        return self


class Always(_Predicate):
    """A predicate that always evaluates to true.

    For example:

    >>> assert True == Always() is True
    >>> assert False == Always() is True
    >>> assert {} == Always() is True
    >>> assert None == Always() is True
    >>> assert 42 == Always() is True
    """

    def __call__(self) -> Self:
        return self

    def __eq__(self, other: Any) -> bool:
        return True

    def __bool__(self) -> bool:
        return True

    def __repr__(self) -> str:
        return "Always()"


Always = Always()  # type: ignore


class Never(_Predicate):
    """A predicate that always evaluates to false.

    For example:

    >>> assert True == Never() is False
    >>> assert False == Never() is False
    >>> assert {} == Never() is False
    >>> assert None == Never() is False
    >>> assert 42 == Never() is False
    """

    def __eq__(self, other: Any) -> bool:
        return False

    def __bool__(self) -> bool:
        return False

    def __repr__(self) -> str:
        return "Never()"


Never = Never()  # type: ignore


class OneOf(_Predicate):
    """A predicate that match one of its values.

    For example:

    >>> assert 1 == OneOf(1, 2, 3)
    >>> assert 2 == OneOf(1, 2, 3)
    >>> assert 3 == OneOf(1, 2, 3)
    >>> assert 4 != OneOf(1, 2, 3)
    """

    values: Collection[Any]

    def __init__(self, *values: Any) -> None:
        self.values = values

    def __eq__(self, other: Any) -> bool:
        return any(value == other for value in self.values)

    def __bool__(self) -> bool:
        raise TypeError("OneOf cannot be inferred to boolean")

    def __repr__(self) -> str:
        return "OneOf()"


class contains:
    """Operand used to check what contains a mapping.

    Example:

    >>> assert {"foo": 42, "bar": True} == contains({"foo": 42})
    >>> assert {"foo": 42, "bar": True} != contains({"qux": "other"})
    """

    def __init__(self, data: Mapping[Any, Any]) -> None:
        self.data: Mapping[Any, Any] = data

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Mapping):
            raise TypeError("Omit object is not comparable")
        return self.data.items() <= other.items()

    def __lt__(self, other: Any) -> bool:
        raise TypeError("Operator < is not supported by contains")

    def __le__(self, other: Any) -> bool:
        raise TypeError("Operator <= is not supported by contains")

    def __gt__(self, other: Any) -> bool:
        raise TypeError("Operator > is not supported by contains")

    def __ge__(self, other: Any) -> bool:
        raise TypeError("Operator >= is not supported by contains")



from __future__ import annotations

import dataclasses
from collections.abc import Callable
from contextlib import suppress
from functools import cached_property
from typing import Any, ClassVar, Generic, Protocol, TypeGuard, TypeVar, cast

T = TypeVar("T")


@dataclasses.dataclass(kw_only=True, slots=True)
class Implementation(Generic[T]):
    is_type: Callable[[Any], TypeGuard[T]]
    get_fields: Callable[[T], set[str]]
    replace: Callable[[T, dict[str, Any]], T]


implementations: list[Implementation] = []


class DataclassInstance(Protocol):
    __dataclass_fields__: ClassVar[dict[str, dataclasses.Field[Any]]]


implementations.append(
    Implementation[DataclassInstance](
        is_type=lambda x: dataclasses.is_dataclass(x) and not isinstance(x, type),  # type: ignore
        get_fields=lambda x: {field.name for field in dataclasses.fields(x)},
        replace=lambda x, changes: dataclasses.replace(x, **changes),
    )
)


implementations.append(
    Implementation[dict](
        is_type=lambda x: isinstance(x, dict), # type: ignore
        get_fields=lambda x: set(x),
        replace=lambda x, changes: x | changes,
    )
)

with suppress(ImportError):
    import attrs

    implementations.append(
        Implementation[attrs.AttrsInstance](
            is_type=lambda x: attrs.has(x) and not isinstance(x, type),  # type: ignore
            get_fields=lambda x: {field.name for field in attrs.fields(type(x))},
            replace=lambda x, changes: attrs.evolve(x, **changes),  # type: ignore
        )
    )


with suppress(ImportError):
    import msgspec

    implementations.append(
        Implementation[msgspec.Struct](
            is_type=lambda x: isinstance(x, msgspec.Struct),  # type: ignore
            get_fields=lambda x: {field.name for field in msgspec.structs.fields(x)},
            replace=lambda x, changes: msgspec.structs.replace(x, **changes),
        )
    )


class Patch:
    pass


class Omit(Patch):
    attrs: frozenset[str]
    __match_args__ = ("attrs",)

    def __init__(self, *attrs: str) -> None:
        self.attrs = frozenset(attrs)

    def __ror__(self, other):
        if isinstance(other, Patch):
            return compose_patch(self, other)

        return Patched(other, self)


class Only(Patch):
    attrs: frozenset[str]
    __match_args__ = ("attrs",)

    def __init__(self, *attrs: str) -> None:
        self.attrs = frozenset(attrs)

    def __ror__(self, other):
        if isinstance(other, Patch):
            return compose_patch(self, other)

        return Patched(other, self)


class Patched(Generic[T]):
    obj: T
    patch: Patch

    def __init__(self, obj: T, patch: Patch) -> None:
        if isinstance(obj, type):
            raise ValueError
        if isinstance(obj, Patched):
            patch, obj = (
                compose_patch(cast(Patched, obj).patch, patch),
                cast(Patched, obj).obj,
            )
        self.obj = obj
        self.patch = patch

    @cached_property
    def wrapped(self) -> Any:
        from . import Always

        obj = self.obj
        for implem in implementations:
            if implem.is_type(obj):
                match self.patch:
                    case Omit(attrs):
                        changes = dict.fromkeys(attrs, Always())
                        return implem.replace(obj, changes)
                    case Only(attrs):
                        all_fields = implem.get_fields(obj)
                        changes = dict.fromkeys(all_fields - attrs, Always())

                        return implem.replace(obj, changes)
                    case _:
                        raise NotImplementedError
        else:
            raise NotImplementedError(f"Not implemented for {type(self.obj)}")

    def __eq__(self, other: Any) -> bool:
        return other == self.wrapped


def compose_patch(lhs: Patch, rhs: Patch) -> Patch:
    match [lhs, rhs]:
        case Omit(attrs=omit_attrs), Only(attrs=only_attrs):
            attrs = only_attrs - omit_attrs
            return Only(*list(attrs))
        case Only(attrs=only_attrs), Omit(attrs=omit_attrs):
            attrs = only_attrs - omit_attrs
            return Only(*list(attrs))
        case Only(attrs=attrs_x), Only(attrs=attrs_y):
            attrs = attrs_x & attrs_y
            return Only(*list(attrs))
        case Omit(attrs=attrs_x), Omit(attrs=attrs_y):
            attrs = attrs_x | attrs_y
            return Omit(*list(attrs))
    raise ValueError



from typing import Any, Mapping

import pytest
from _pytest.assertion.util import assertrepr_compare

import worth


def pytest_assertrepr_compare(
    config: pytest.Config, op: str, left: Any, right: Any
) -> list[str] | None:
    if isinstance(left, worth._Predicate):
        return [f"{right!r} never match"]
    if isinstance(right, worth._Predicate):
        return [f"{left!r} never match"]

    patched = False
    if isinstance(left, worth.Patched):
        patched = True
        left = left.wrapped
    if isinstance(right, worth.Patched):
        patched = True
        right = right.wrapped

    if patched:
        return assertrepr_compare(config, op, left, right)

    contained = False
    if isinstance(left, Mapping) and isinstance(right, worth.contains):
        contained = True
        right = dict(left) | dict(right.data)
    if isinstance(left, worth.contains) and isinstance(right, Mapping):
        contained = True
        left = dict(right) | dict(left.data)

    if contained:
        message = assertrepr_compare(config, op, left, right) or []
        if op == "==":
            message.insert(0, f"{left} contains {right}")
        elif op == "!=":
            message.insert(0, f"{left} does not contain {right}")
        return message

    return None
